﻿using UnityEngine;
using System.Collections.Generic;
using System.Collections;
using UnityEngine.SceneManagement;
using UnityEngine.UI;
using HarmonyLib;
using Steamworks;

public class BenevolentSprites : Mod
{
    public static List<HelperSprite> sprites;
    static GameObject firePrefab;
    public static Item_Base gardenItem;
    public static Item_Base animalItem;
    public static AI_NetworkBehaviour_Domestic_Resource[] animals;
    public static PickupItem_Networked[] eggs;
    public static BeeHive[] beehives;
    Harmony harmony;
    static float updateTime = 0;
    public static List<Message_InitiateConnection> waiting;
    public static bool gameLoaded;
    public static string fileName;
    public static GameObject FirePrefab
    {
        get
        {
            if (firePrefab == null)
                foreach (ParticleSystemRenderer p in Resources.FindObjectsOfTypeAll<ParticleSystemRenderer>())
                    if (p.name == "Fire_v2")
                    { firePrefab = p.gameObject; break; }
            return firePrefab;
        }
    }
    public void Start()
    {
        fileName = modlistEntry.modinfo.modFile.Name;
        gameLoaded = false;
        waiting = new List<Message_InitiateConnection>();
        animals = null;
        eggs = null;
        if (SceneManager.GetActiveScene().name != Semih_Network.MenuSceneName)
        {
            modlistEntry.modinfo.unloadBtn.GetComponent<Button>().onClick.Invoke();
            throw new ModLoadException("Mod must be loaded on the main menu");
        }
        var Plank = ItemManager.GetItemByName("Plank");
        gardenItem = Plank.Clone(8765, "Garden_Sprite");
        gardenItem.settings_Inventory.LocalizationTerm = "";
        gardenItem.settings_Inventory.DisplayName = "Garden Sprite";
        gardenItem.settings_Inventory.Description = "A little glowing friend.\nIt seems to take interest in nearby plants.";
        Traverse.Create(gardenItem.settings_Inventory).Field("stackSize").SetValue(1);
        var size = Plank.settings_Inventory.Sprite.rect.size;
        gardenItem.settings_Inventory.Sprite = Sprite.Create(new Texture2D((int)size.x, (int)size.y, TextureFormat.RGBA32, false), new Rect(0, 0, size.x, size.y), new Vector2(0, 0), Plank.settings_Inventory.Sprite.pixelsPerUnit);
        int width = (int)size.x;
        int height = (int)size.y;
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++)
                gardenItem.settings_Inventory.Sprite.texture.SetPixel(x, y, new Color(0.2f, 1, 0, 1 - new Vector2(x - width / 2, y - height / 2).magnitude / width * 2));
        gardenItem.settings_Inventory.Sprite.texture.Apply();


        animalItem = gardenItem.Clone(8766, "Animal_Sprite");
        animalItem.settings_Inventory.DisplayName = "Animal Sprite";
        animalItem.settings_Inventory.Description = "A little glowing friend.\nIt seems to take interest in nearby animals.";
        animalItem.settings_Inventory.Sprite = Sprite.Create(new Texture2D((int)size.x, (int)size.y, TextureFormat.RGBA32, false), new Rect(0, 0, size.x, size.y), new Vector2(0, 0), Plank.settings_Inventory.Sprite.pixelsPerUnit);
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++)
                animalItem.settings_Inventory.Sprite.texture.SetPixel(x, y, new Color(1, 0, 1, 1 - new Vector2(x - width / 2, y - height / 2).magnitude / width * 2));
        animalItem.settings_Inventory.Sprite.texture.Apply();

        modlistEntry.jsonmodinfo.isModPermanent = true;
        HMLLibrary.ModManagerPage.RefreshModState(modlistEntry);

        RAPI.RegisterItem(gardenItem);
        RAPI.RegisterItem(animalItem);
        foreach (SO_RandomDropper rand in Resources.FindObjectsOfTypeAll<SO_RandomDropper>())
            if (rand.name.Contains("Dropper_FishingRod_Metal"))
            {
                var items = new List<RandomItem>(rand.randomizer.items);
                items.Add(new RandomItem() { obj = gardenItem, weight = 1.5f, spawnChance = "" });
                items.Add(new RandomItem() { obj = animalItem, weight = 1.5f, spawnChance = "" });
                rand.randomizer.items = items.ToArray();
                var total = rand.randomizer.TotalWeight;
                foreach (RandomItem item in rand.randomizer.items)
                    item.spawnChance = (item.weight / total * 100) + "%";
            }

        sprites = new List<HelperSprite>();
        harmony = new Harmony("com.aidanamite.BenevolentSprites");
        harmony.PatchAll();
        Log("Mod has been loaded!");
    }

    public static HelperSprite CreateSprite(Storage_Small box, Slot item, uint SpriteIndex = 0)
    {
        HelperSprite newFire = null;
        if (item.itemInstance.UniqueIndex == gardenItem.UniqueIndex)
            newFire = new GardenSprite(box, item, SpriteIndex);
        else if (item.itemInstance.UniqueIndex == animalItem.UniqueIndex)
            newFire = new AnimalSprite(box, item, SpriteIndex);
        sprites.Add(newFire);
        if (Semih_Network.IsHost)
            new Message_Sprite_Create(newFire).Message.Broadcast();
        return newFire;
    }
    public static HelperSprite CreateSprite(Message_InitiateConnection message)
    {
        if (message.appBuildID == MessageType.Create)
        {
            var data = new Message_Sprite_Create(message);
            return CreateSprite(data.Box, data.Slot, data.SpriteIndex);
        }
        else if (message.appBuildID == MessageType.Recreate)
        {
            var data = new Message_Sprite_Recreate(message);
            var sprite = CreateSprite(data.Box, data.Slot, data.SpriteIndex);
            sprite.gameObject.transform.localPosition = data.Position;
            sprite.SetTarget(sprite.FindTarget(data.TargetIndex).transform, data.Offset);
            return sprite;
        }
        return null;
    }
    public static HelperSprite CreateSprite(string serial)
    {
        var data = serial.Bytes();
        var sprite = CreateSprite(new Message_InitiateConnection(0, MessageType.Recreate, serial));
        sprite.holding.AddRange(data.Items(out int inc, 40));
        List<byte> extra = new List<byte>();
        for (int i = 40 + inc; i < data.Length; i++)
            extra.Add(data[i]);
        sprite.RestoreData(extra.ToArray());
        sprite.store.itemInstance.exclusiveString = "";
        return sprite;
    }

    public void Update()
    {
        for (int i = sprites.Count - 1; i >= 0; i--)
            if (sprites[i].IsDead)
                sprites.RemoveAt(i);
            else
                sprites[i].Update(Time.deltaTime);
        updateTime += Time.deltaTime;
        if (updateTime > 1)
        {
            updateTime %= 1;
            eggs = NetworkIDManager.GetNetworkdIDs<PickupItem_Networked>(NetworkIDTag.ChickenEgg);
            animals = FindObjectsOfType<AI_NetworkBehaviour_Domestic_Resource>();
            beehives = NetworkIDManager.GetNetworkdIDs<BeeHive>();
        }
        var message = RAPI.ListenForNetworkMessagesOnChannel(MessageType.ChannelID);
        if (message != null && message.message != null && message.message is Message_InitiateConnection && message.message.Type == MessageType.MessageID)
        {
            var msg = (Message_InitiateConnection)message.message;
            if (gameLoaded)
                ProcessMessage(msg);
            else
                waiting.Add(msg);
        }
    }

    public static void ProcessMessage(Message_InitiateConnection msg)
    {
        if (msg.appBuildID == MessageType.Create || msg.appBuildID == MessageType.Recreate)
            CreateSprite(msg);
        else if (msg.appBuildID == MessageType.Target)
            Message_Sprite_SetTarget.Use(msg);
        else if (msg.appBuildID == MessageType.Kill)
            Message_Sprite_Kill.Use(msg);
    }

    public override void WorldEvent_OnPlayerConnected(CSteamID steamid, RGD_Settings_Character characterSettings)
    {
        if (Semih_Network.IsHost)
            foreach (HelperSprite sprite in sprites)
                new Message_Sprite_Recreate(sprite).Message.Send(steamid);
    }
    public override void WorldEvent_WorldLoaded()
    {
        if (Semih_Network.IsHost)
        {
            gameLoaded = true;
            foreach (Storage_Small box in StorageManager.allStorages)
                foreach (Slot slot in box.GetInventoryReference().allSlots)
                    if (slot.HasSprite())
                    {
                        if (slot.itemInstance.exclusiveString == "")
                            CreateSprite(box, slot);
                        else
                            CreateSprite(slot.itemInstance.exclusiveString);
                    }
        }
    }

    public override void WorldEvent_WorldUnloaded()
    {
        gameLoaded = false;
    }
}

static class ExtentionMethods
{
    static Dictionary<Slot, Storage_Small> refs = new Dictionary<Slot, Storage_Small>();
    public static Sprite CreateEmpty(this Sprite sprite)
    {
        return Sprite.Create(new Texture2D(sprite.texture.width, sprite.texture.height,TextureFormat.RGBA32,false), sprite.rect, sprite.pivot, sprite.pixelsPerUnit);
    }
    public static Item_Base Clone(this Item_Base source, int uniqueIndex, string uniqueName)
    {
        Item_Base item = ScriptableObject.CreateInstance<Item_Base>();
        item.Initialize(uniqueIndex, uniqueName, source.MaxUses);
        item.settings_buildable = source.settings_buildable.Clone();
        item.settings_consumeable = source.settings_consumeable.Clone();
        item.settings_cookable = source.settings_cookable.Clone();
        item.settings_equipment = source.settings_equipment.Clone();
        item.settings_Inventory = source.settings_Inventory.Clone();
        item.settings_recipe = source.settings_recipe.Clone();
        item.settings_usable = source.settings_usable.Clone();
        return item;
    }

    public static bool HasSprite(this Slot slot)
    {
        return slot.HasValidItemInstance() && (slot.itemInstance.UniqueIndex == BenevolentSprites.gardenItem.UniqueIndex || slot.itemInstance.UniqueIndex == BenevolentSprites.animalItem.UniqueIndex);
    }

    public static HelperSprite GetSprite(this Slot slot)
    {
        foreach (HelperSprite sprite in BenevolentSprites.sprites)
            if (sprite.store == slot)
                return sprite;
        return null;
    }

    public static Storage_Small GetStorage(this Slot slot)
    {
        if (refs.ContainsKey(slot))
        {
            if (refs[slot] == null)
            {
                var inv = Traverse.Create(slot).Field("inventory").GetValue<Inventory>();
                foreach (Storage_Small storage in StorageManager.allStorages)
                    if (storage.GetInventoryReference() == inv)
                    {
                        refs[slot] = storage;
                        return storage;
                    }
                }
            return refs[slot];
        } else
        {
            var inv = Traverse.Create(slot).Field("inventory").GetValue<Inventory>();
            foreach (Storage_Small storage in StorageManager.allStorages)
                if (storage.GetInventoryReference() == inv)
                {
                    refs.Add(slot,storage);
                    return storage;
                }
        }
        return null;
    }

    public static void Close(this Storage_Small box)
    {
        new Message_Storage_Close(Messages.StorageManager_Close, RAPI.GetLocalPlayer().StorageManager, box).Broadcast(NetworkChannel.Channel_Game);
    }

    public static List<ItemInstance> GetAllItems(this PickupItem pickup)
    {
        var items = new List<ItemInstance>();
        if (pickup.yieldHandler != null)
            items.AddAll(pickup.yieldHandler.Yield);
        if (pickup.dropper != null)
            items.AddAll(pickup.dropper.GetRandomItems());
        if (pickup.itemInstance != null && pickup.itemInstance.Valid)
            items.Add(pickup.itemInstance);
        if (pickup.specificPickups != null)
            foreach (Pickup_Specific specificPickup in pickup.specificPickups)
            {
                var pickupInternal = Traverse.Create(specificPickup);
                foreach (string name in pickupInternal.Fields())
                {
                    object value = pickupInternal.Field(name).GetValue();
                    if (value is RandomDropper)
                        items.AddAll((value as RandomDropper).GetRandomItems());
                }
            }
        return items;
    }

    public static void AddAll(this List<ItemInstance> items, IEnumerable<Cost> newItems)
    {
        if (newItems != null)
            foreach (Cost cost in newItems)
                if (cost != null)
                    items.Add(new ItemInstance(cost.item, cost.amount, cost.item.MaxUses));
    }
    public static void AddAll(this List<ItemInstance> items, IEnumerable<Item_Base> newItems)
    {
        if (newItems != null)
            foreach (Item_Base item in newItems)
                if (item != null)
                    items.Add(new ItemInstance(item, 1, item.MaxUses));
    }
    public static void AddAll(this List<ItemInstance> items, IEnumerable<ItemInstance> newItems)
    {
        if (newItems != null)
            foreach (ItemInstance item in newItems)
                if (item != null)
                    items.Add(item);
    }
    public static void Remove(this PickupItem_Networked pickup)
    {
        if (pickup.stopTrackUseRPC)
        {
            PickupObjectManager.RemovePickupItemNetwork(pickup, new CSteamID(0));
        }
        else
        {
            PickupObjectManager.RemovePickupItem(pickup, new CSteamID(0));
        }
    }
    public static string String(this byte[] bytes, int length = -1, int offset = 0)
    {
        string str = "";
        if (length == -1)
            length = (bytes.Length - offset) / 2;
        while (str.Length < length)
        {
            str += System.BitConverter.ToChar(bytes,offset + str.Length * 2);
        }
        return str;
        
    }
    public static string String(this List<byte> bytes) => bytes.ToArray().String();
    public static byte[] Bytes(this string str) {
        var data = new List<byte>();
        foreach (char chr in str)
            data.AddRange(System.BitConverter.GetBytes(chr));
        return data.ToArray();
    }
    public static int Integer(this byte[] bytes, int offset = 0) => System.BitConverter.ToInt32(bytes, offset);
    public static uint UInteger(this byte[] bytes, int offset = 0) => System.BitConverter.ToUInt32(bytes, offset);
    public static float Float(this byte[] bytes, int offset = 0) => System.BitConverter.ToSingle(bytes, offset);
    public static Vector3 Vector3(this byte[] bytes, int offset = 0) => new Vector3(bytes.Float(offset), bytes.Float(offset + 4), bytes.Float(offset + 8));
    public static ItemInstance Item(this byte[] bytes, out int size, int offset = 0)
    {
        var stringSize = bytes.Integer(offset + 12);
        size = stringSize / 2 + 16;
        return new ItemInstance(ItemManager.GetItemByIndex(bytes.Integer(offset)), bytes.Integer(offset + 4), bytes.Integer(offset + 8), bytes.String(stringSize, offset + 16));
    }
    public static List<ItemInstance> Items(this byte[] bytes, out int size, int offset = 0)
    {
        var items = new List<ItemInstance>();
        var count = bytes.Integer(offset);
        size = 4;
        while (items.Count < count)
        {
            items.Add(bytes.Item(out int inc, offset + size));
            size += inc;
        }
        return items;
    }
    public static byte[] Bytes(this int value) => System.BitConverter.GetBytes(value);
    public static byte[] Bytes(this uint value) => System.BitConverter.GetBytes(value);
    public static byte[] Bytes(this float value) => System.BitConverter.GetBytes(value);
    public static byte[] Bytes(this Vector3 value)
    {
        var data = new byte[12];
        value.x.Bytes().CopyTo(data, 0);
        value.y.Bytes().CopyTo(data, 4);
        value.z.Bytes().CopyTo(data, 8);
        return data;
    }
    public static byte[] Bytes(this ItemInstance value)
    {
        var data = new List<byte>();
        data.AddRange(value.baseItem.UniqueIndex.Bytes());
        data.AddRange(value.Amount.Bytes());
        data.AddRange(value.Uses.Bytes());
        data.AddRange(value.exclusiveString.Length.Bytes());
        data.AddRange(value.exclusiveString.Bytes());
        return data.ToArray();
    }
    public static byte[] Bytes(this IEnumerable<ItemInstance> values)
    {
        var data = new List<byte>();
        int count = 0;
        foreach (ItemInstance item in values)
        {
            data.AddRange(item.Bytes());
            count++;
        }
        data.InsertRange(0, count.Bytes());
        return data.ToArray();
    }

    public static void Broadcast(this Message message, NetworkChannel channel = (NetworkChannel)MessageType.ChannelID) => ComponentManager<Semih_Network>.Value.RPC(message, Target.Other, EP2PSend.k_EP2PSendReliable, channel);
    public static void Send(this Message message, CSteamID steamID) => ComponentManager<Semih_Network>.Value.SendP2P(steamID, message, EP2PSend.k_EP2PSendReliable, (NetworkChannel)MessageType.ChannelID);
}

public class HelperSprite
{
    internal GameObject gameObject;
    internal Slot store;
    public List<ItemInstance> holding;
    const float acceleration = 2f;
    const float maxVelocity = 10f;
    Vector3 velocity;
    internal Transform targetObject;
    public Vector3 targetPosition;
    Vector3 randomAccel;
    Vector3 targetRand;
    float time;
    internal bool dying;
    public uint ObjectIndex { private set; get; }
    internal Storage_Small box
    {
        get
        {
            return store.GetStorage();
        }
    }
    public bool IsDead
    {
        get { return gameObject == null; }
    }
    public Vector3 Target
    {
        get
        {
            if (targetObject == null)
                return targetPosition;
            return gameObject.transform.parent.InverseTransformPoint(targetObject.TransformPoint(targetPosition));
        }
    }
    public Vector3 TargetGlobal
    {
        get
        {
            return gameObject.transform.parent.TransformPoint(Target);
        }
    }
    internal float DistanceFromTarget { get { return (Target - gameObject.transform.localPosition).magnitude; } }
    public HelperSprite(Storage_Small container, Slot slot, Color color, uint objectIndex = 0)
    {
        gameObject = GameObject.Instantiate(BenevolentSprites.FirePrefab, GameManager.Singleton.lockedPivot, false);
        if (objectIndex == 0)
            objectIndex = SaveAndLoad.GetUniqueObjectIndex();
        ObjectIndex = objectIndex;
        store = slot;
        gameObject.name = "sprite";
        gameObject.transform.position = container.transform.position;
        gameObject.transform.localScale = Vector3.one / 2;
        foreach (ParticleSystem p in gameObject.GetComponentsInChildren<ParticleSystem>())
        {
            var m = p.main;
            m.startColor = new ParticleSystem.MinMaxGradient(color);
        }
        Light glow = gameObject.AddComponent<Light>();
        glow.color = color;
        glow.intensity = 1;
        glow.range = 2;
        glow.enabled = true;
        gameObject.SetActive(true);
        SetTarget(container.transform, Vector3.up);
        randomAccel = Vector3.zero;
        dying = false;
        holding = new List<ItemInstance>();
    }
    public void SetTarget(Transform target, Vector3 position)
    {
        targetObject = target;
        targetPosition = position;
    }
    public void SetTargetNetwork(MonoBehaviour_ID target, Vector3 position)
    {
        SetTarget(target.transform, position);
        if (Semih_Network.IsHost)
            new Message_Sprite_SetTarget(this, target, position).Message.Broadcast();
    }

    public void Update(float t)
    {
        if (store == null || !store.HasSprite())
            Kill();
        time += t;
        if (time > 0.2)
        {
            targetRand = new Vector3(Random.Range(-1f, 1f), Random.Range(-1f, 1f), Random.Range(-1f, 1f));
            time -= 0.2f;
        }
        randomAccel = Vector3.Lerp(randomAccel, targetRand, t / 2);
        Vector3 dir = Target + randomAccel - gameObject.transform.localPosition;
        if (!dying)
            velocity += dir.normalized * acceleration * t;
        velocity *= 0.98f;
        velocity.Clamp(0, maxVelocity);
        gameObject.transform.localPosition += velocity * t;
        AI();
    }

    public virtual void Kill()
    {
        if (!dying)
        {
            if (Semih_Network.IsHost)
                new Message_Sprite_Kill(this).Message.Broadcast();
            CoroutineManager.Singleton.StartCoroutine(Death());
        }
    }

    IEnumerator Death()
    {
        dying = true;
        float timePassed = 0;
        Light light = gameObject.GetComponent<Light>();
        foreach (ItemInstance item in holding)
            Helper.DropItem(item, gameObject.transform.position, Vector3.down, true);
        while (timePassed < 0.75)
        {
            timePassed += Time.deltaTime / 2;
            float strength = Mathf.Max(Mathf.Sin(timePassed * Mathf.PI) * 3, 0);
            light.intensity = strength;
            light.range = strength * 2;
            yield return new WaitForEndOfFrame();
        }
        Destroy();
        yield break;
    }

    public void Destroy()
    {
        GameObject.Destroy(gameObject);
    }

    void AI()
    {
        if (!Semih_Network.IsHost || dying)
            return;
        if (tryInteract())
        {
            MonoBehaviour_ID target = null;
            Vector3 targetOffset = Vector3.zero;
            if (holding.Count == 0 && targetObject == box.transform)
                pickTarget(out target, out targetOffset);
            if (target == null)
            {
                target = box;
                if (holding.Count == 0)
                    targetOffset = Vector3.up;
                else
                    targetOffset = Vector3.up * 0.2f;
            }
            if ((target == null ? null : target.transform) != targetObject || targetOffset != targetPosition)
                SetTargetNetwork(target, targetOffset);
        }
    }

    internal bool tryDepositHeld()
    {
        if (targetObject == box.transform && DistanceFromTarget < 0.2f && !box.IsOpen)
        {
            bool added = false;
            for (int i = holding.Count - 1; i >= 0; i--)
            {
                int pre = holding[i].Amount;
                box.GetInventoryReference().AddItem(holding[i], false);
                if (pre != holding[i].Amount)
                    added = true;
                if (holding[i].Amount == 0)
                    holding.RemoveAt(i);
            }
            if (added)
            {
                var eventRef = Traverse.Create(ComponentManager<SoundManager>.Value).Field("eventRef_UI_MoveItem").GetValue<string>();
                box.Close();
                var msg = new Message_SoundManager_PlayOneShot(Messages.SoundManager_PlayOneShot, ComponentManager<Semih_Network>.Value.NetworkIDManager, ComponentManager<SoundManager>.Value.ObjectIndex, eventRef, box.transform.position);
                msg.Broadcast();
                FMODUnity.RuntimeManager.PlayOneShot(eventRef, msg.Position);
            }
        }
        return holding.Count == 0;
    }

    internal virtual bool tryInteract()
    {
        return true;
    }

    public virtual void RestoreData(byte[] data) { }
    public virtual byte[] GenerateData() { return new byte[0]; }

    internal virtual void pickTarget(out MonoBehaviour_ID target, out Vector3 targetOffset) { target = box; targetOffset = Vector3.up; }

    public virtual MonoBehaviour_ID FindTarget(uint objectIndex)
    {
        foreach (var mono in GameObject.FindObjectsOfType<MonoBehaviour_ID>())
            if (mono.ObjectIndex == objectIndex)
                return mono;
        return null;
    }

}

class GardenSprite : HelperSprite
{
    static Dictionary<GardenSprite, PlantationSlot> targeted = new Dictionary<GardenSprite, PlantationSlot>();
    public GardenSprite(Storage_Small container, Slot slot, uint objectIndex = 0) : base(container, slot, new Color(Random.Range(0, 0.125f), Random.Range(0.875f, 1), 0), objectIndex)
    {
        targeted.Add(this, null);
    }

    public override void Kill()
    {
        targeted.Remove(this);
        base.Kill();
    }

    internal override bool tryInteract()
    {
        if (targetObject == null)
            return true;
        if (targetObject == box.transform)
            return tryDepositHeld();
        if (targeted[this].plant == null || !targeted[this].plant.FullyGrown())
        {
            targeted[this] = null;
            return true;
        }
        if (DistanceFromTarget > 0.2f)
            return false;
        holding.AddRange(targeted[this].plant.pickupComponent.GetAllItems());
        var seed = targeted[this].plant.item.UniqueIndex;
        targetObject.GetComponent<Cropplot>().plantManager.Harvest(targeted[this].plant, false);
        new Message_HarvestPlant(Messages.PlantManager_HarvestPlant, targetObject.GetComponent<Cropplot>().plantManager, targeted[this].plant, false).Broadcast(NetworkChannel.Channel_Game);
        foreach (ItemInstance item in holding)
            if (item.baseItem.UniqueIndex == seed)
            {
                PlantSeed(targetObject.GetComponent<Cropplot>(), seed);
                if (item.Amount == 1)
                    holding.Remove(item);
                else
                    item.Amount--;
                break;
            }
        targeted[this] = null;
        return true;
    }
    internal override void pickTarget(out MonoBehaviour_ID target, out Vector3 targetOffset)
    {
        float dist = float.MaxValue;
        target = null;
        targetOffset = Vector3.zero;
        targeted[this] = null;
        foreach (var cropplot in PlantManager.allCropplots)
        {
            if (cropplot is Cropplot_Grass)
                continue;
            foreach (var plantation in cropplot.plantationSlots)
            {
                if (targeted.ContainsValue(plantation))
                    continue;
                var newDist = Vector3.Distance(box.transform.position, plantation.transform.position);
                if (plantation.plant != null && plantation.plant.FullyGrown() && newDist < dist)
                {
                    target = cropplot;
                    targetOffset = plantation.plant.transform.localPosition + Vector3.up * 0.2f;
                    dist = newDist;
                    targeted[this] = plantation;
                }
            }
        }
        if (target != null)
        {
            if (dist > box.GetInventoryReference().GetSlotCount() / 2)
            {
                target = null;
                targetOffset = Vector3.zero;
                targeted[this] = null;
            } else 
            if (!box.IsOpen && box.GetInventoryReference().GetItemCount(targeted[this].plant.item) > 0)
            {
                box.GetInventoryReference().RemoveItem(targeted[this].plant.item.UniqueName, 1);
                box.Close();
                holding.Add(new ItemInstance(targeted[this].plant.item, 1, targeted[this].plant.item.MaxUses));
            }
        }
        else
            targeted[this] = null;
    }

    static void PlantSeed(Cropplot crop, int ItemIndex)
    {
        var network = ComponentManager<Semih_Network>.Value;
        var plantPrefab = crop.plantManager.GetPlantByIndex(ItemIndex);
        var msg = new Message_PlantSeed(Messages.PlantManager_PlantSeed, crop.plantManager, crop, plantPrefab, false);
        if (Semih_Network.IsHost)
        {
            msg.plantObjectIndex = SaveAndLoad.GetUniqueObjectIndex();
            msg.waterPlantedSeed = ComponentManager<WeatherManager>.Value.GetCurrentWeatherType() == UniqueWeatherType.Rain;
            crop.plantManager.PlantSeed(crop, plantPrefab, msg.plantObjectIndex, msg.waterPlantedSeed, false);
            network.RPC(msg, global::Target.Other, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
            return;
        }
        msg.plantObjectIndex = 0U;
        network.SendP2P(network.HostID, msg, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
        return;
    }

    public override void RestoreData(byte[] data)
    {
        if (targetObject.GetComponent<Cropplot>() != null)
            targeted[this] = targetObject.GetComponent<Cropplot>().plantationSlots[data.Integer(0)];
    }
    public override byte[] GenerateData()
    {
        if (targeted[this] != null)
            return targetObject.GetComponent<Cropplot>().plantationSlots.IndexOf(targeted[this]).Bytes();
        return new byte[0];
    }

    public override MonoBehaviour_ID FindTarget(uint objectIndex)
    {
        foreach (var cropplot in PlantManager.allCropplots)
            if (cropplot.ObjectIndex == objectIndex)
                return cropplot;
        return base.FindTarget(objectIndex);
    }
}

class AnimalSprite : HelperSprite
{
    static Dictionary<AnimalSprite, MonoBehaviour_ID> targeted = new Dictionary<AnimalSprite, MonoBehaviour_ID>();
    public AnimalSprite(Storage_Small container, Slot slot, uint objectIndex = 0) : base(container, slot, new Color(Random.Range(0.125f, 0.25f), 0, Random.Range(0.875f, 1)), objectIndex)
    {
        targeted.Add(this, null);
    }
    public override void Kill()
    {
        targeted.Remove(this);
        base.Kill();
    }

    internal override void pickTarget(out MonoBehaviour_ID target, out Vector3 targetOffset)
    {
        target = null;
        targetOffset = Vector3.zero;
        float dist = float.MaxValue;
        targeted[this] = null;
        Item_Base requiredItem = null;
        var itemCount = new Dictionary<Item_Base, int>();
        if (BenevolentSprites.animals != null)
            foreach (AI_NetworkBehaviour_Domestic_Resource animal in BenevolentSprites.animals)
            {
                if (animal == null || targeted.ContainsValue(animal))
                    continue;
                var newDist = Vector3.Distance(box.transform.position, animal.transform.position);
                var collectionItem = animal.Resource.resource.collectionItem;
                itemCount.TryAdd(collectionItem, box.GetInventoryReference().GetItemCount(collectionItem));
                if (animal.Resource != null && animal.Resource.IsReady && newDist < dist && (collectionItem.MaxUses != 1 || (!box.IsOpen && itemCount[collectionItem] > 0)))
                {
                    if (collectionItem.MaxUses == 1)
                        requiredItem = collectionItem;
                    else
                        requiredItem = null;
                    target = animal;
                    targetOffset = Vector3.up * 0.2f;
                    dist = newDist;
                    targeted[this] = animal;
                }
            }
        if (BenevolentSprites.eggs != null)
            foreach (PickupItem_Networked item in BenevolentSprites.eggs)
            {
                if (item == null || targeted.ContainsValue(item))
                    continue;
                var newDist = Vector3.Distance(box.transform.position, item.transform.position);
                if (newDist < dist)
                {
                    requiredItem = null;
                    target = item;
                    targetOffset = Vector3.up * 0.2f;
                    dist = newDist;
                    targeted[this] = item;
                }
            }
        if (BenevolentSprites.beehives != null)
            foreach (BeeHive hive in BenevolentSprites.beehives)
            {
                if (hive == null || targeted.ContainsValue(hive))
                    continue;
                var newDist = Vector3.Distance(box.transform.position, hive.transform.position);
                if (hive.currentHoneyLevelIndex != -1 && newDist < dist)
                {
                    requiredItem = null;
                    target = hive;
                    targetOffset = Vector3.up * 0.2f;
                    dist = newDist;
                    targeted[this] = hive;
                }
            }
        foreach (BirdsNest nest in BirdsNest.AllNests)
        {
            if (targeted.ContainsValue(nest))
                continue;
            var newDist = Vector3.Distance(box.transform.position, nest.transform.position);
            if ((nest.HasFeather || nest.HasEgg) && newDist < dist)
            {
                requiredItem = null;
                target = nest;
                targetOffset = Vector3.up * 0.2f;
                dist = newDist;
                targeted[this] = nest;
            }
        }
        if (target != null)
        {
            if (dist > box.GetInventoryReference().GetSlotCount() / 2)
            {
                target = null;
                targetOffset = Vector3.zero;
                targeted[this] = null;
            }
            else if (requiredItem != null)
            {
                box.GetInventoryReference().RemoveItemUses(requiredItem.UniqueName, 1);
                box.Close();
                holding.Add(new ItemInstance(requiredItem, 1, 1));
            }
        }
    }
    internal override bool tryInteract()
    {
        if (targetObject == null)
            return true;
        if (targetObject == box.transform)
            return tryDepositHeld();
        ResourceRegenerative resource = null;
        var animal = targetObject.GetComponent<AI_NetworkBehaviour_Domestic_Resource>();
        var pickup = targetObject.GetComponent<PickupItem_Networked>();
        var beehive = targetObject.GetComponent<BeeHive>();
        var nest = targetObject.GetComponent<BirdsNest>();
        if (animal != null)
        {
            resource = animal.Resource;
            if (!resource.IsReady)
            {
                targeted[this] = null;
                return true;
            }
        }
        if (beehive != null && beehive.currentHoneyLevelIndex == -1)
        {
            targeted[this] = null;
            return true;
        }
        if (nest != null && !nest.HasEgg && !nest.HasFeather)
        {
            targeted[this] = null;
            return true;
        }
        if (DistanceFromTarget > 0.2f)
            return false;
        if (animal != null)
        {
            Item_Base RequiredItem = resource.resource.collectionItem;
            Item_Base _pickup = resource.HarvestResource();
            holding.Add(new ItemInstance(_pickup,1,_pickup.MaxUses));
            if (RequiredItem.MaxUses == 1)
                foreach (ItemInstance item in holding)
                    if (item.baseItem.UniqueIndex == RequiredItem.UniqueIndex)
                    {
                        if (item.Amount == 1)
                            holding.Remove(item);
                        else
                            item.Amount--;
                        break;
                    }
        }
        else if (beehive != null)
        {
            holding.AddAll(beehive.CurrentHoneyLevel.yield.yieldAssets);
            new Message_NetworkBehaviour_ID_SteamID(Messages.BeeHiveHarvestOutput, ComponentManager<Semih_Network>.Value.NetworkIDManager, beehive.ObjectIndex, new CSteamID(0)).Broadcast();
            Traverse.Create(beehive).Method("HarvestYield", new System.Type[] { typeof(Network_Player) }, new object[] { null }).GetValue();
        }
        else if (nest != null)
        {
            var item = Traverse.Create(nest).Field("eggItem").GetValue<Item_Base>();
            if (nest.EggCount > 0)
            {
                holding.Add(new ItemInstance(item, nest.EggCount, item.MaxUses));
                new Message_NetworkBehaviour_SteamID(Messages.BirdsNest_PickupEgg, nest, 0).Broadcast();
                nest.PickupEgg(null);
            }
            item = Traverse.Create(nest).Field("featherItem").GetValue<Item_Base>();
            if (nest.FeatherCount > 0)
            {
                holding.Add(new ItemInstance(item, nest.FeatherCount, item.MaxUses));
                new Message_NetworkBehaviour_SteamID(Messages.BirdsNest_PickupFeather, nest, 0).Broadcast();
                nest.PickupFeather(null);
            }
        }
        else
        {
            holding.AddRange(pickup.PickupItem.GetAllItems());
            pickup.Remove();
        }
        targeted[this] = null;
        return true;
    }

    public override void RestoreData(byte[] data)
    {
        targeted[this] = targetObject.GetComponent<MonoBehaviour_ID>();
    }

    public override MonoBehaviour_ID FindTarget(uint objectIndex)
    {
        if (BenevolentSprites.animals != null)
            foreach (var animal in BenevolentSprites.animals)
                if (animal.ObjectIndex == objectIndex)
                    return animal;
        if (BenevolentSprites.eggs != null)
            foreach (var item in BenevolentSprites.eggs)
                if (item.ObjectIndex == objectIndex)
                    return item;
        if (BenevolentSprites.beehives != null)
            foreach (var hive in BenevolentSprites.beehives)
                if (hive.ObjectIndex == objectIndex)
                    return hive;
        foreach (var nest in BirdsNest.AllNests)
            if (nest.ObjectIndex == objectIndex)
                return nest;
        return base.FindTarget(objectIndex);
    }
}

class ModLoadException : System.Exception
{
    public ModLoadException(string message) : base(message) { }
}

[HarmonyPatch(typeof(HMLLibrary.ModManagerPage), "ShowModInfo")]
class Patch_ShowModInfo
{
    static void Postfix(ModData md)
    {
        if (md.modinfo.mainClass != null && md.modinfo.mainClass.GetType() == typeof(BenevolentSprites))
            HMLLibrary.ModManagerPage.modInfoObj.transform.Find("MakePermanent").gameObject.SetActive(true);
    }
}

[HarmonyPatch(typeof(Storage_Small), "Close")]
class Patch_CloseStorage
{
    static void Postfix(Storage_Small __instance)
    {
        if (BenevolentSprites.gameLoaded && Semih_Network.IsHost)
        {
            foreach (Slot slot in __instance.GetInventoryReference().allSlots)
                if (slot.HasSprite() && slot.GetSprite() == null)
                {
                    if (slot.itemInstance.exclusiveString != "")
                        BenevolentSprites.CreateSprite(slot.itemInstance.exclusiveString);
                    else
                        BenevolentSprites.CreateSprite(__instance, slot);
                }
        }
    }
}

[HarmonyPatch(typeof(GameManager), "OnWorldRecievedLate")]
class Patch_RemoteWorldLoaded
{
    static void Postfix(Storage_Small __instance)
    {
        BenevolentSprites.gameLoaded = true;
        foreach (Message_InitiateConnection message in BenevolentSprites.waiting)
            BenevolentSprites.ProcessMessage(message);
        BenevolentSprites.waiting.Clear();
    }
}

[HarmonyPatch(typeof(SaveAndLoad), "SaveWorld")]
class Patch_SaveWorld
{
    static void Prefix()
    {
        foreach (HelperSprite sprite in BenevolentSprites.sprites)
            sprite.store.itemInstance.exclusiveString = new Message_Sprite_Recreate(sprite).Message.password + sprite.holding.Bytes().String() + sprite.GenerateData().String();
    }
    static void Postfix()
    {
        foreach (HelperSprite sprite in BenevolentSprites.sprites)
            sprite.store.itemInstance.exclusiveString = "";
    }
}

[HarmonyPatch(typeof(SoundManager), "PlayUI_MoveItem")]
class Patch_PlayMoveItemSound
{
    static bool Prefix()
    {
        return !System.Environment.StackTrace.Contains("HelperSprite.tryDepositHeld");
    }
}

public struct MessageType
{
    public const int ChannelID = 8;
    public const Messages MessageID = (Messages)23468;
    public const int Create = 0;
    public const int Target = 1;
    public const int Kill = 2;
    public const int Recreate = 3;
}

class Message_Sprite_Create
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(spriteIndex.Bytes());
            data.AddRange(boxIndex.Bytes());
            data.AddRange(slotIndex.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Create, data.String());
        }
    }
    public Storage_Small Box
    {
        get
        {
            foreach (Storage_Small box in StorageManager.allStorages)
                if (box.ObjectIndex == boxIndex)
                    return box;
            return null;
        }
    }
    public Slot Slot
    {
        get
        {
            var box = Box;
            if (box == null || box.GetInventoryReference().allSlots.Count <= slotIndex || slotIndex < 0)
                return null;
            return box.GetInventoryReference().allSlots[slotIndex];
        }
    }
    public HelperSprite Sprite
    {
        get
        {
            foreach (HelperSprite sprite in BenevolentSprites.sprites)
                if (sprite.ObjectIndex == spriteIndex)
                    return sprite;
            return null;
        }
    }
    public uint SpriteIndex { get { return spriteIndex; } }
    uint spriteIndex;
    uint boxIndex;
    int slotIndex;
    public Message_Sprite_Create(HelperSprite sprite)
    {
        spriteIndex = sprite.ObjectIndex;
        boxIndex = sprite.box.ObjectIndex;
        slotIndex = sprite.box.GetInventoryReference().allSlots.IndexOf(sprite.store);
    }
    public Message_Sprite_Create(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        spriteIndex = data.UInteger(0);
        boxIndex = data.UInteger(4);
        slotIndex = data.Integer(8);
    }
}

class Message_Sprite_SetTarget
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(spriteIndex.Bytes());
            data.AddRange(targetIndex.Bytes());
            data.AddRange(targetOffset.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Target, data.String());
        }
    }
    public Vector3 Offset { get { return targetOffset; } }
    public MonoBehaviour_ID Target
    {
        get
        {
            return Sprite.FindTarget(targetIndex);
        }
    }
    public HelperSprite Sprite
    {
        get
        {
            foreach (HelperSprite sprite in BenevolentSprites.sprites)
                if (sprite.ObjectIndex == spriteIndex)
                    return sprite;
            return null;
        }
    }
    uint spriteIndex;
    uint targetIndex;
    Vector3 targetOffset;
    public Message_Sprite_SetTarget(HelperSprite sprite,MonoBehaviour_ID Target,Vector3 TargetOffset)
    {
        spriteIndex = sprite.ObjectIndex;
        targetIndex = Target.ObjectIndex;
        targetOffset = TargetOffset;
    }
    Message_Sprite_SetTarget(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        spriteIndex = data.UInteger(0);
        targetIndex = data.UInteger(4);
        targetOffset = data.Vector3(8);
    }

    public static void Use(Message_InitiateConnection message)
    {
        var msg = new Message_Sprite_SetTarget(message);
        var target = msg.Target;
        var sprite = msg.Sprite;
        if (target == null)
            return;
        if (sprite == null)
            return;
        sprite.SetTarget(target.transform, msg.Offset);
    }
}

class Message_Sprite_Kill
{
    public Message_InitiateConnection Message
    {
        get
        {
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Kill, spriteIndex.Bytes().String());
        }
    }
    public HelperSprite Sprite
    {
        get
        {
            foreach (HelperSprite sprite in BenevolentSprites.sprites)
                if (sprite.ObjectIndex == spriteIndex)
                    return sprite;
            return null;
        }
    }
    uint spriteIndex;
    public Message_Sprite_Kill(HelperSprite sprite)
    {
        spriteIndex = sprite.ObjectIndex;
    }
    Message_Sprite_Kill(Message_InitiateConnection message)
    {
        spriteIndex = message.password.Bytes().UInteger(0);
    }

    public static void Use(Message_InitiateConnection message) => new Message_Sprite_Kill(message).Sprite.Kill();
}

class Message_Sprite_Recreate
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(spriteIndex.Bytes());
            data.AddRange(boxIndex.Bytes());
            data.AddRange(slotIndex.Bytes());
            data.AddRange(targetIndex.Bytes());
            data.AddRange(targetOffset.Bytes());
            data.AddRange(position.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Recreate, data.String());
        }
    }
    public Storage_Small Box
    {
        get
        {
            foreach (Storage_Small box in StorageManager.allStorages)
                if (box.ObjectIndex == boxIndex)
                    return box;
            return null;
        }
    }
    public Slot Slot
    {
        get
        {
            var box = Box;
            if (box == null || box.GetInventoryReference().allSlots.Count <= slotIndex || slotIndex < 0)
                return null;
            return box.GetInventoryReference().allSlots[slotIndex];
        }
    }
    public HelperSprite Sprite
    {
        get
        {
            foreach (HelperSprite sprite in BenevolentSprites.sprites)
                if (sprite.ObjectIndex == spriteIndex)
                    return sprite;
            return null;
        }
    }
    public Vector3 Offset { get { return targetOffset; } }
    public Vector3 Position { get { return position; } }
    public uint TargetIndex { get { return targetIndex; } }
    public uint SpriteIndex { get { return spriteIndex; } }
    uint spriteIndex;
    uint boxIndex;
    int slotIndex;
    uint targetIndex;
    Vector3 targetOffset;
    Vector3 position;
    public Message_Sprite_Recreate(HelperSprite sprite)
    {
        spriteIndex = sprite.ObjectIndex;
        boxIndex = sprite.box.ObjectIndex;
        slotIndex = sprite.box.GetInventoryReference().allSlots.IndexOf(sprite.store);
        targetIndex = sprite.targetObject.GetComponent<MonoBehaviour_ID>().ObjectIndex;
        targetOffset = sprite.targetPosition;
        position = sprite.gameObject.transform.localPosition;
    }
    public Message_Sprite_Recreate(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        spriteIndex = data.UInteger(0);
        boxIndex = data.UInteger(4);
        slotIndex = data.Integer(8);
        targetIndex = data.UInteger(12);
        targetOffset = data.Vector3(16);
        position = data.Vector3(28);
    }
}